/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

// Functions for working with RSA and Rabin keys. The intent is that the 
// rest of the DQM code be able to use a uniform interface to these two
// different key types.
#include <openssl/pem.h>
#include <openssl/err.h>
#include "sha1.h"
#include "async.h"
#include "cryptkey.h"
#include "crypt.h"

// all public keys are stored in a format like:
// [rsa|rab]<KEY>
// the first three bytes identifies the key type. the rest of the KEY
// is described below.
ptr<pubkey>
str_to_pubkey(const str& s)
{
    const char* p = s.cstr();
    if (!memcmp(p, "rab", 3)) {
	return rabin_pubkey::from_str(&p[3], s.len() - 3);

    } else if (!memcmp(p, "rsa", 3)) {
	return rsa_pubkey::from_str(&p[3], s.len() - 3);

    } else {
	// special case handling so we are backwards
	// compatible with the RSA keys that don't have the prefix.
	// XXX: might want to fix later. slight chance of bug -- 
	// there might be an RSA key that actually starts out,
	// in base 64, with 'rab'.
	return rsa_pubkey::from_str(p, s.len());
    }
}

ptr<pubkey>
file_to_pubkey(const str& fname)
{
    str s = file2str(fname);
    if (s.cstr() == NULL) return NULL;

    // get rid of any newlines
    if (s[s.len() - 1] == '\n') 
	s.setbuf(s.cstr(), s.len() - 1);

    return str_to_pubkey(s);
}

// This function tries to figure out which kind of private key is stored
// in the file given by 'fname'. Right now, the "figuring out" is quite
// crude. Note that public keys, unlike private keys, need the same format
// because they are passed around "in-band". Private keys, in our
// application, always hang out on the disk and so they don't need to be
// stored in the same format. Obviously, if we use this code for another
// application, we'll need to revisit.
ptr<privkey>
file_to_privkey(const str& fname)
{
    str s = file2str(fname);
    if (s.cstr() == NULL) return NULL;

    const char* p = s.cstr();
    static const char* rsa_header = "-----BEGIN RSA PRIVATE KEY-----";

    // hack.
    if (!memcmp(p, rsa_header, strlen(rsa_header))) {

	FILE* f = fopen(fname, "r");
	assert(f);
	if (f == NULL) return NULL;

	RSA* rsa = PEM_read_RSAPrivateKey(f, NULL, NULL, NULL);
	fclose(f);

	if (rsa == NULL) return NULL;

	return New refcounted<rsa_privkey> (rsa);

    } else {
	 
	// just assume Rabin here

	if (s[s.len() - 1] == '\n')  
	    s = substr(s, 0, s.len() - 1);

	return rabin_privkey::from_str(s);
    } 
}


// Rabin

// Rabin public keys are stored as base64-encoded binary strings. the
// binary looks like:
// <n-in-raw-form>
ptr<rabin_pubkey>
rabin_pubkey::from_str(const char* p, int len)
{
    str s = dearmor64(p, len);
    if (s.cstr() == NULL) return NULL;

    bigint n;
    n.setraw(s);

    ptr<rabin_pub> pubk = New refcounted<rabin_pub> (n);

    return New refcounted<rabin_pubkey>(pubk);
}

ptr<rabin_privkey> 
gen_rabin_privkey(int nbits)
{
    rabin_priv privk = rabin_keygen(nbits);
    ptr<rabin_priv> privk_copy = New refcounted<rabin_priv>(privk.p, privk.q);
    return New refcounted<rabin_privkey> (privk_copy);
}

ptr<rsa_privkey>   
gen_rsa_privkey(int nbits, unsigned long e)
{
    static bool seeded = false;
    if (!seeded) {
	srand(time(NULL));
	seeded = true;
    }

    RSA* rsa = RSA_generate_key(nbits, e, NULL, NULL);
    if (rsa == NULL) {
	warnx("error generating key. error: %s\n", 
	      ERR_error_string(ERR_get_error(), NULL));
	return NULL;
    }

    return New refcounted<rsa_privkey>(rsa);
}

str
rabin_pubkey::to_str() const
{
    strbuf sb;
    sb << "rab" << armor64(pubk->n.getraw());
    return sb;
}

bool
rabin_pubkey::verify(const str& msg, const str& sig) const
{
    bigint sig_bi;

    sig_bi.setraw(sig);

    return pubk->verify(msg, sig_bi);
}

str
rabin_privkey::sign(const str& msg) const
{
    bigint sig = privk->sign(msg);
    return sig.getraw();
}

// Rabin private keys are stored as base-64 encoded binary strings. The
// binary format is:
// <4-byte-big-endian-length-of-p><raw p><4-byte-big-endian-length-of-q><raw q>
ptr<rabin_privkey>
rabin_privkey::from_str(const str& orig_str)
{
    str s = dearmor64(orig_str.cstr(), orig_str.len());
    if (s.cstr() == NULL) return NULL;

    bigint p, q;

    const char* pq_str = s.cstr();
    int len_p = getint(pq_str);
    pq_str = &pq_str[4];
    p.setraw(str(pq_str, len_p));

    pq_str = &pq_str[len_p];
    int len_q = getint(pq_str);
    pq_str = &pq_str[4];
    q.setraw(str(pq_str, len_q));

    ptr<rabin_priv> privk = New refcounted<rabin_priv>(p, q);
    return New refcounted<rabin_privkey>(privk);
}


str
rabin_privkey::to_str() const
{
    str p_str = privk->p.getraw();
    str q_str = privk->q.getraw();
    
    int p_len, q_len;
    
    p_len = p_str.len();
    q_len = q_str.len();

    mstr m(4 + p_len + 4 + q_len);

    char* mp = m.cstr();

    putint(mp, p_len);
    mp = &mp[4];

    memcpy(mp, p_str.cstr(), p_len);
    mp = &mp[p_len];

    putint(mp, q_len);
    mp = &mp[4];

    memcpy(mp, q_str.cstr(), q_len);

    return armor64(m);
}

// RSA

// RSA public keys are stored as base64-encoded binary strings. the binary 
// looks like:
// <4-byte-big-endian-length-of-n><n><4-byte-big-endian-length-of-e><e>
ptr<rsa_pubkey>
rsa_pubkey::from_str(const char* p, int len)
{
    str s = dearmor64(p, len);
    if (s.cstr() == NULL) return NULL;

    BIGNUM* bn_n = NULL;
    BIGNUM* bn_e = NULL;

    const char* n_str = s.cstr();
    int len_n = getint(n_str) + 4;
    if ( (bn_n = BN_mpi2bn(reinterpret_cast<const uint8_t*>(n_str), 
                           len_n, NULL)) == NULL) 
	return NULL;

    const char* e_str = &n_str[len_n];
    int len_e = getint(e_str) + 4;

    if ( (bn_e = BN_mpi2bn(reinterpret_cast<const uint8_t*>(e_str), 
			   len_e, NULL)) == NULL) {
	BN_free(bn_n);
	return NULL;
    }

    RSA* rsa = RSA_new();
    rsa->n = bn_n;
    rsa->e = bn_e;

    return New refcounted<rsa_pubkey>(rsa);
}

str
rsa_pubkey::to_str() const
{
    // figure out how long n and e will be when stringized
    int n_len = BN_bn2mpi(rsa->n, NULL);
    int e_len = BN_bn2mpi(rsa->e, NULL);

    // allocate space for the flattened n and e
    mstr m(n_len + e_len);

    // now do the flattening
    BN_bn2mpi(rsa->n, reinterpret_cast<uint8_t*>(m.cstr()));
    BN_bn2mpi(rsa->e, reinterpret_cast<uint8_t*>(m.cstr()) + n_len);

    strbuf sb;
    sb << "rsa" << armor64(m);

    return sb;
}

// this function is directly copied from DM.
// XXX: put DM's copyright notice in somewhere.
inline void
sha1oracle_lookup (int idx, char *dst, size_t dstlen,
                   const char *src, size_t srclen)
{
    sha1oracle ora (dstlen, idx);
    ora.update (src, srclen);
    ora.final (reinterpret_cast<u_char *> (dst));
}

const int hindex = 36;
// The padding scheme we use for sign and verify is Full-Domain-Hash --
//  we use DM's sha1oracle to construct a message digest that is the same
//  size as the signature (also the same number of bits as the RSA
//  modulus). (this is in contrast to using SHA1 as the digest-producer;
//  the output of SHA1 is much smaller than the size of the signature).
//  Signing consists of taking the output of a particular sha1_oracle and
//  finding its RSA inverse.
//  Verifying is taking the signature, raising it to the e power mod N, and
//  then checking that the output of that function is equal to the output
//  of the same sha1_oracle.


// Pre-condition: digest must point to 'digestlen' bytes of memory. 
// 'digestlen' needs to be the length of the RSA modulus in bytes. 
// Returns false if something breaks.
// For k bit moduli, this function returns a (k-1) bit digest; it simply
// zeroes the top bit of the oracle and moves all the other bits downward.
bool
get_digest(mstr& digest, int digestlen, const str& msg, RSA* rsa)
{
    if (digestlen != RSA_size(rsa)) {
	assert(false);
	return false;
    }

    bool ret = false;
    BIGNUM *oracle_res_bn = NULL; 
    BIGNUM *digest_bn     = NULL;
    int bitshift = 1, 
        top_bit  = digestlen*8 - 1;

    int num_bytes, zeroes_needed;
    uint8_t* digeststr = reinterpret_cast<uint8_t*>(digest.cstr());
    const uint8_t* const_digeststr = digeststr;

    sha1oracle_lookup(hindex, digest.cstr(), digestlen, msg.cstr(), msg.len());

    // turn oracle output into BIGNUM
    if ((oracle_res_bn = BN_bin2bn(const_digeststr, digestlen, NULL)) == NULL)
	goto err;

    if ((digest_bn = BN_new()) == NULL)
	goto err;

    // this is to handle keys whose lengths aren't multiples of 8. in that
    // case, rsa_size gives a number in bytes, but we need to shift
    // downward so that the digest is smaller than the modulus. (the top
    // byte of the modulus might start with some number of zeroes)
    while (!BN_is_bit_set(rsa->n, top_bit)) {
	--top_bit;
	++bitshift;
    }

    // digest = oracle_res >> bitshift
    if (!BN_rshift(digest_bn, oracle_res_bn, bitshift))
	goto err;

    // now get the digest into the passed-in storage
    num_bytes     = BN_num_bytes(digest_bn);
    zeroes_needed = digestlen - num_bytes;
    assert(zeroes_needed >= 0);

    if (zeroes_needed == 0)
	BN_bn2bin(digest_bn, digeststr);
    else {
	BN_bn2bin(digest_bn, digeststr + zeroes_needed); 
	memset(digest.cstr(), 0, zeroes_needed);
    }
    ret = true;

err:
    if (digest_bn)       BN_free(digest_bn);
    if (oracle_res_bn)   BN_free(oracle_res_bn);

    return ret;
}

bool
rsa_pubkey::verify(const str& msg, const str& sig) const
{
    int siglen = sig.len();

    if (siglen != RSA_size(rsa))
	return false;

    mstr decrypted_sig(siglen);
    mstr msg_digest(siglen);

    uint8_t* sig_ptr = reinterpret_cast<uint8_t*>(
			         const_cast<char *>(sig.cstr()));

    if (RSA_public_decrypt(siglen, 
                           sig_ptr, 
		           reinterpret_cast<uint8_t*>(decrypted_sig.cstr()),
		           rsa,
		           RSA_NO_PADDING) < 0) 
	return false;

    if (!get_digest(msg_digest, siglen, msg, rsa))
	return false;
    
    return !memcmp(msg_digest.cstr(), decrypted_sig, siglen);
}


str
rsa_privkey::sign(const str& msg) const
{
    int siglen = RSA_size(rsa);
    mstr sig(siglen);
    // digest is the same length as the signature (that's the point of FDH)
    mstr digest(siglen);

    if (!get_digest(digest, siglen, msg, rsa))
	return NULL;

    if (RSA_private_encrypt(siglen,
                            reinterpret_cast<uint8_t*>(digest.cstr()), 
		            reinterpret_cast<uint8_t*>(sig.cstr()), 
			    rsa,
		            RSA_NO_PADDING) < 0)  {
	printf("RSA error: %s\n", ERR_error_string(ERR_get_error(), NULL));
	return NULL;
    }

    sig.setlen(siglen);
    return sig;
}

bool
rsa_privkey::to_file(const str& fname) const
{
    // the private key file shouldn't be readable or writable by anyone but
    // the user.
    mode_t oldumask = umask(0077);
    FILE* f = fopen(fname, "w");
    umask(oldumask);
    if (f == NULL) return false;

    int rc = PEM_write_RSAPrivateKey(f, rsa, NULL, NULL, 0, NULL, NULL);
    fclose(f);

    return rc;
}
