/*
 * Copyright (C) 2005 J.D. Zamfirescu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

// This code wraps C++ pubkey and privkey objects so that Python can
// use them.
#include <Python.h>
#include "structmember.h"
#include "cryptkey.h"

typedef struct {
  PyObject_HEAD
  ptr<pubkey> p;
} PublicKey;

typedef struct {
  PyObject_HEAD
  ptr<privkey> p;
} PrivateKey;

static void
PublicKey_dealloc(PublicKey* self)
{
  self->p = NULL;
  self->ob_type->tp_free((PyObject*)self);
}

static void
PrivateKey_dealloc(PrivateKey* self)
{
  self->p = NULL;
  self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
PublicKey_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PublicKey *self;

  self = (PublicKey *)type->tp_alloc(type, 0);
  if (self != NULL) {
    self->p = NULL;
  }

  return (PyObject *)self;
}

static PyObject *
PrivateKey_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PrivateKey *self;

  self = (PrivateKey *)type->tp_alloc(type, 0);
  if (self != NULL) {
    self->p = NULL;
  }

  return (PyObject *)self;
}



static PyMemberDef PublicKey_members[] = {
  {NULL}  /* Sentinel */
};

static PyMemberDef PrivateKey_members[] = {
  {NULL}  /* Sentinel */
};

static PyObject *
PublicKey_verify(PublicKey *self, PyObject *args)
{
  int ok;
  const char *msg, *sig;
  int sig_len;
  int msg_len;
  bool v;

  ok = PyArg_ParseTuple(args, "s#s#", &msg, &msg_len, &sig, &sig_len);
  if (! ok)
    return Py_BuildValue("");

  const str msgs = str(msg, msg_len);
  const str sigs = str(sig, sig_len);

  v = self->p->verify(msgs, sigs);

  // MW: do we have to worry about Py_BuildValue possibly returning a NULL
  // pointer? (no, we don't -- zamfi)
  return Py_BuildValue("i", v);
} 

static PyObject *
PublicKey_to_str(PublicKey *self)
{
  assert(self->p != NULL);
  str s = self->p->to_str();

  const char *ns = s.cstr();
  if (ns == NULL)
    return Py_BuildValue("");

  return Py_BuildValue("s", ns);
}

static PyObject *
PrivateKey_to_str(PrivateKey *self)
{
  assert(self->p != NULL);
  str s = self->p->to_str();

  const char *ns = s.cstr();
  if (ns == NULL)
    return Py_BuildValue("");

  return Py_BuildValue("s", ns);
}


static PyObject *
PublicKey_to_file(PublicKey *self, PyObject *args)
{
  int ok;
  const char *path;

  assert(self->p != NULL);

  ok = PyArg_ParseTuple(args, "s", &path);
  if (! ok)
    return Py_BuildValue("");
  const str paths = str(path);

  bool v = self->p->to_file(paths);
  return Py_BuildValue("i", v);
}


static PyObject *
PrivateKey_to_file(PrivateKey *self, PyObject *args)
{
  int ok;
  const char *path;
  
  ok = PyArg_ParseTuple(args, "s", &path);
  if (! ok)
    return Py_BuildValue("");

  assert(self->p != NULL);

  const str paths = str(path);
  bool v = self->p->to_file(paths);
  return Py_BuildValue("i", v);
}

static PyObject *
PrivateKey_sign(PrivateKey *self, PyObject *args)
{
  int ok;
  const char *msg;
  int len;
  str v;

  ok = PyArg_ParseTuple(args, "s#", &msg, &len);
  if (! ok)
    return Py_BuildValue("");

  const str msgs = str(msg, len); // check for msgs being a valid str?
  if (msgs.cstr() == NULL) return Py_BuildValue("");
  v = self->p->sign(msgs);
  if (v.cstr() == NULL) 
    return Py_BuildValue("");

  return Py_BuildValue("s#", v.cstr(), v.len());
}

static PyObject *
PrivateKey_to_pubkey(PrivateKey *self);

static PyMethodDef PublicKey_methods[] = {
  {"verify", (PyCFunction)PublicKey_verify, METH_VARARGS,
   "Verify the message and signature using the internal private key"
  },
  {"to_str", (PyCFunction)PublicKey_to_str, METH_NOARGS,
   "Return a string representation of the key"
  },
  {"to_file", (PyCFunction)PublicKey_to_file, METH_VARARGS,
   "Create a file of the key"
  },
  {NULL}  /* Sentinel */
};

static PyMethodDef PrivateKey_methods[] = {
  {"sign", (PyCFunction)PrivateKey_sign, METH_VARARGS,
   "Sign the argument using the internal private key"
  },
  {"to_str", (PyCFunction)PrivateKey_to_str, METH_NOARGS,
   "Return a string representation of the key"
  },
  {"to_file", (PyCFunction)PrivateKey_to_file, METH_VARARGS,
   "Create a file of the key"
  },
  {"to_pubkey", (PyCFunction)PrivateKey_to_pubkey, METH_NOARGS,
   "Create the corresponding public key for this private key"
  },
  {NULL}  /* Sentinel */
};


static PyTypeObject PublicKeyType = {
  PyObject_HEAD_INIT(NULL)
  0,                         /*ob_size*/
  "cryptkey.PublicKey",             /*tp_name*/
  sizeof(PublicKey),             /*tp_basicsize*/
  0,                         /*tp_itemsize*/
  (destructor)PublicKey_dealloc, /*tp_dealloc*/
  0,                         /*tp_print*/
  0,                         /*tp_getattr*/
  0,                         /*tp_setattr*/
  0,                         /*tp_compare*/
  0,                         /*tp_repr*/
  0,                         /*tp_as_number*/
  0,                         /*tp_as_sequence*/
  0,                         /*tp_as_mapping*/
  0,                         /*tp_hash */
  0,                         /*tp_call*/
  0,                         /*tp_str*/
  0,                         /*tp_getattro*/
  0,                         /*tp_setattro*/
  0,                         /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "PublicKey objects",           /* tp_doc */
  0,		               /* tp_traverse */
  0,		               /* tp_clear */
  0,		               /* tp_richcompare */
  0,		               /* tp_weaklistoffset */
  0,		               /* tp_iter */
  0,		               /* tp_iternext */
  PublicKey_methods,             /* tp_methods */
  PublicKey_members,             /* tp_members */
  0,                         /* tp_getset */
  0,                         /* tp_base */
  0,                         /* tp_dict */
  0,                         /* tp_descr_get */
  0,                         /* tp_descr_set */
  0,                         /* tp_dictoffset */
  0,                         /* tp_init */
  0,                         /* tp_alloc */
  0,                         /* tp_new */
};

static PyTypeObject PrivateKeyType = {
  PyObject_HEAD_INIT(NULL)
  0,                         /*ob_size*/
  "cryptkey.PrivateKey",             /*tp_name*/
  sizeof(PrivateKey),             /*tp_basicsize*/
  0,                         /*tp_itemsize*/
  (destructor)PrivateKey_dealloc, /*tp_dealloc*/
  0,                         /*tp_print*/
  0,                         /*tp_getattr*/
  0,                         /*tp_setattr*/
  0,                         /*tp_compare*/
  0,                         /*tp_repr*/
  0,                         /*tp_as_number*/
  0,                         /*tp_as_sequence*/
  0,                         /*tp_as_mapping*/
  0,                         /*tp_hash */
  0,                         /*tp_call*/
  0,                         /*tp_str*/
  0,                         /*tp_getattro*/
  0,                         /*tp_setattro*/
  0,                         /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "PrivateKey objects",           /* tp_doc */
  0,		               /* tp_traverse */
  0,		               /* tp_clear */
  0,		               /* tp_richcompare */
  0,		               /* tp_weaklistoffset */
  0,		               /* tp_iter */
  0,		               /* tp_iternext */
  PrivateKey_methods,             /* tp_methods */
  PrivateKey_members,             /* tp_members */
  0,                         /* tp_getset */
  0,                         /* tp_base */
  0,                         /* tp_dict */
  0,                         /* tp_descr_get */
  0,                         /* tp_descr_set */
  0,                         /* tp_dictoffset */
  0,                         /* tp_init */
  0,                         /* tp_alloc */
  0,                         /* tp_new */
};

static PyObject *
PrivateKey_to_pubkey(PrivateKey *self)
{
  ptr<pubkey> pkey;

  pkey = self->p->to_pubkey();

  if (pkey == NULL)
    return Py_BuildValue("");

  PyObject *pkey_py = PublicKey_new(&PublicKeyType, 0, 0);
  if (pkey_py == NULL)
    return Py_BuildValue("");
  // MW: what if pkey_py is NULL? this comment applies to all of the
  // invocations of PublicKey_new
  ((PublicKey *)pkey_py)->p = pkey;

  return pkey_py;
}


static PyObject *
CK_file_to_pubkey(PyObject *self, PyObject *args) {
  int ok;
  const char *s;
  ptr<pubkey> pkey;

  ok = PyArg_ParseTuple(args, "s", &s);
  if (! ok)
    return Py_BuildValue("");

  const str keys = str(s);

  pkey = file_to_pubkey(keys);
  if (pkey == NULL)
    return Py_BuildValue("");

  PyObject *pkey_py = PublicKey_new(&PublicKeyType, 0, 0);
  if (pkey_py == NULL)
    return Py_BuildValue("");
  ((PublicKey *)pkey_py)->p = pkey;

  return pkey_py;
}

static PyObject *
CK_file_to_privkey(PyObject *self, PyObject *args) {
  int ok;
  const char *s;
  ptr<privkey> pkey;

  ok = PyArg_ParseTuple(args, "s", &s);
  if (! ok)
    return Py_BuildValue("");

  const str keys = str(s);

  pkey = file_to_privkey(keys);
  if (pkey == NULL)
    return Py_BuildValue("");

  PyObject *pkey_py = PrivateKey_new(&PrivateKeyType, 0, 0);
  if (pkey_py == NULL)
    return Py_BuildValue("");
  // MW: same comment as with PublicKey_new. what if pkey_py is NULL?
  ((PrivateKey *)pkey_py)->p = pkey;

  return pkey_py;
}

static PyObject *
CK_str_to_pubkey(PyObject *self, PyObject *args) {
  int ok;
  const char *s;
  ptr<pubkey> pkey;

  ok = PyArg_ParseTuple(args, "s", &s);
  if (! ok)
    return Py_BuildValue("");

  const str keys = str(s);

  pkey = str_to_pubkey(keys);
  if (pkey == NULL)
    return Py_BuildValue("");

  PyObject *pkey_py = PublicKey_new(&PublicKeyType, 0, 0);
  if (pkey_py == NULL)
    return Py_BuildValue("");
  ((PublicKey *)pkey_py)->p = pkey;

  return pkey_py;
}

static PyObject *
CK_gen_rabin_privkey(PyObject *self, PyObject *args) {
  int ok; 
  int nbits;
  ptr<privkey> pkey;
  
  ok = PyArg_ParseTuple(args, "i", &nbits);
  if (! ok)
    return Py_BuildValue("");

  pkey = gen_rabin_privkey(nbits);
  if (pkey == NULL)
    return Py_BuildValue("");

  PyObject *pkey_py = PrivateKey_new(&PrivateKeyType, 0, 0);
  if (pkey_py == NULL)
    return Py_BuildValue("");
  ((PrivateKey *)pkey_py)->p = pkey;

  return pkey_py;
}

static PyObject *
CK_gen_rsa_privkey(PyObject *self, PyObject *args) {
  int ok; 
  int nbits, e;
  ptr<privkey> pkey;
  
  ok = PyArg_ParseTuple(args, "ik", &nbits, &e);
  if (! ok)
    return Py_BuildValue("");
  
  pkey = gen_rsa_privkey(nbits, e);
  if (pkey == NULL)
    return Py_BuildValue("");

  PyObject *pkey_py = PrivateKey_new(&PrivateKeyType, 0, 0);
  if (pkey_py == NULL)
    return Py_BuildValue("");
  ((PrivateKey *)pkey_py)->p = pkey;

  return pkey_py;
}

static PyMethodDef module_methods[] = {
  {"str_to_pubkey", (PyCFunction)CK_str_to_pubkey, METH_VARARGS,
   "Return a public key from a string"},
  {"file_to_pubkey", (PyCFunction)CK_file_to_pubkey, METH_VARARGS,
   "Return a public key from a filename"},
  {"file_to_privkey", (PyCFunction)CK_file_to_privkey, METH_VARARGS,
   "Return a private key from a filename"},
  {"gen_rabin_privkey", (PyCFunction)CK_gen_rabin_privkey, METH_VARARGS,
   "Generate a new Rabin private key"},
  {"gen_rsa_privkey", (PyCFunction)CK_gen_rsa_privkey, METH_VARARGS,
   "Generate a new Rabin private key"},
  {NULL}  /* Sentinel */
};

#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
initcryptkey(void) 
{
  PyObject* m;

  if (PyType_Ready(&PublicKeyType) < 0)
    return;
  if (PyType_Ready(&PrivateKeyType) < 0)
    return;

  m = Py_InitModule3("cryptkey", module_methods,
		     "Example module that creates an extension type.");

  if (m == NULL)
    return;

  Py_INCREF(&PublicKeyType);
  PyModule_AddObject(m, "PublicKey", (PyObject *)&PublicKeyType);
  Py_INCREF(&PrivateKeyType);
  PyModule_AddObject(m, "PrivateKey", (PyObject *)&PrivateKeyType);
}



/********************************************************************
 * stuff removed, kept here for future reference, if we need to 
 * change anything. The template code can be found here:
 * 
 * http://docs.python.org/ext/node22.html
 ********************************************************************/




// static int
// PublicKey_init(PublicKey *self, PyObject *args, PyObject *kwds)
// {
//   PyObject *first=NULL, *last=NULL, *tmp;

//   static char *kwlist[] = {"first", "last", "number", NULL};

//   if (! PyArg_ParseTupleAndKeywords(args, kwds, "|OOi", kwlist, 
// 				    &first, &last, 
// 				    &self->number))
//     return -1; 

//   if (first) {
//     tmp = self->first;
//     Py_INCREF(first);
//     self->first = first;
//     Py_XDECREF(tmp);
//   }

//   if (last) {
//     tmp = self->last;
//     Py_INCREF(last);
//     self->last = last;
//     Py_XDECREF(tmp);
//   }

//   return 0;
// }



// static PyObject *
// PublicKey_name(PublicKey* self)
// {
//   static PyObject *format = NULL;
//   PyObject *args, *result;

//   if (format == NULL) {
//     format = PyString_FromString("%s %s");
//     if (format == NULL)
//       return NULL;
//   }

//   if (self->first == NULL) {
//     PyErr_SetString(PyExc_AttributeError, "first");
//     return NULL;
//   }

//   if (self->last == NULL) {
//     PyErr_SetString(PyExc_AttributeError, "last");
//     return NULL;
//   }

//   args = Py_BuildValue("OO", self->first, self->last);
//   if (args == NULL)
//     return NULL;

//   result = PyString_Format(format, args);
//   Py_DECREF(args);
    
//   return result;
// }

// from members.
//   {"first", T_OBJECT_EX, offsetof(PublicKey, first), 0,
//    "first name"},
//   {"last", T_OBJECT_EX, offsetof(PublicKey, last), 0,
//    "last name"},
//   {"number", T_INT, offsetof(PublicKey, number), 0,
//    "cryptkey number"},
