# 
# $Id: udp_iface.py,v 1.12 2006/05/19 23:50:35 mwalfish Exp $
#
#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import socket
import errno
import dqm.rpc.events as dqm_events
import dqm.logger

# This class represents a bunch of UDP datagrams that are to be sent or
# received. The XDR UDP client and server use these to get packets and to
# send packets. Eventually, we might want to rate limit in both the
# receive and send directions. 
class udp_dgrams:

    # 'local_port' is the port on which to receive packets. normally
    # clients set this to zero. 'initially_reading' should be set to 'True' 
    # if the object is a server. 'packet_handler' is the callback that
    # gets "packet read" events, and 'local_ip' should be set if you want
    # to specify a particular IP address to bind to (for servers). the
    # callback will be called with a list of one or more 'packet_tuples'
    # which are simply tuples in which the first element is the packet
    # buffer and the second is the address of the sender.
    def __init__ (self, local_port, initially_reading, packet_handler,
	          local_ip='', logger=None):

	if logger is None: logger = dqm.logger.stderr_logger()
	self.logger = logger

	self.sock = None

	# store pending to-send packets in parallel lists so we
	# can use the map function
	self.to_send_msg  = [] 
	self.to_send_addr = []

	# pre-allocate for efficiency
	self.rcvd_tuples = [0]*1024

	self.writing = False
	self.reading = False
	self.cb = packet_handler

	try: 
	    self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	    self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1024*1024)
	    self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 1024*1024)
	    if local_port != 0: self.sock.bind( (local_ip, int(local_port)) )
	    self.sock.setblocking(0)

	except (socket.error), val:
	    logmsg = 'socket got error: %s' % str(val)
	    self.logger.log(logmsg)
	    raise ValueError

	else:
	    if (initially_reading):
		dqm_events.reg_fd_cb(self.sock.fileno(), 
				     dqm_events.READ,
				     self.recv_packets)
		self.reading = True

    def set_reading(self, reading):

	if self.reading is False and reading is True:
	    dqm_events.reg_fd_cb(self.sock.fileno(),
				 dqm_events.READ,
				 self.recv_packets)

	elif self.reading is True and reading is False:
	    dqm_events.unreg_fd(self.sock.fileno(), dqm_events.READ)

	self.reading = reading
	    

    def recv_packets(self):
   
	# use local variable for speed
	recvfrom = self.sock.recvfrom	
	keep_going = True
	rcvd_tuples = self.rcvd_tuples
	i = 0
	while keep_going and i < 1024:

	    try:    
		rcvd_tuples[i] = recvfrom(1024)
		i = i + 1

	    except (socket.error), val:
		if val[0] not in [errno.EAGAIN, errno.EINTR]:
		    logmsg = 'got error code: %s' % errno.errorcode[val[0]]
		    self.logger.log(logmsg)
		keep_going = False

	if i: self.cb(rcvd_tuples,i)
	   

    def send_packets(self):

	# local variables, for speed
	sendto = self.sock.sendto
	to_send_msg = self.to_send_msg
	to_send_addr = self.to_send_addr
	# map, for speed
	map(sendto, to_send_msg, to_send_addr)
	to_send_msg[:] = []
	to_send_addr[:] = []
	dqm_events.unreg_fd(self.sock.fileno(), dqm_events.WRITE)
        self.writing = False


    def queue_packet(self, msg, addr):
	""" 'msg' is the message to send. It will be stuck in a UDP
	          datagram and sent to 'addr'.
	    'addr' is an (IP,port) pair. 'IP' is a string and can be a
	          domain name. 'port' is numeric, in host byte order."""
#	    'cb', if not None, will be called when the network request 
#	          has completed. The argument to 'cb' is a tuple, t:
#		 --> If t[0] is True, the request completed successfully.
#		 --> If t[0] is False, the request was unsuccessful and
#		       t[1] is a tuple giving the error number and cause.
#	"""

	self.to_send_msg.append(msg)
	self.to_send_addr.append(addr)
	if not self.writing:
	    dqm_events.reg_fd_cb(self.sock.fileno(), 
			         dqm_events.WRITE, 
				 self.send_packets)
	    self.writing = True
	

