#
# $Id: rcvr.py,v 1.6 2006/05/19 23:50:35 mwalfish Exp $
# 
# Top-level functions executed by stamp recipients.
# 
#
#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import os
import os.path
import time
import sha
import dqm.rpc.events     as dqm_events
import dqm.crypt.stamp    as dqm_stamp
import dqm.user.dqm_client as dqm_client
import dqm.err            as dqm_err
import dqm.work_queue.q   as work_q
import dqm.logger

#
# stamp, logical version:
#   Version
#   Cert 
#   Disposable
#
# stamp, stringized version:
#   version#signer_id:pub_key:quota:expiry:sig#i:weekstamp:sig 
#    <all binary values are base64 encoded>
#
# DQM val:
#    hash(stamp_stringized)
#
# DQM key:
#   postmark = hash(DQM val)


def init_set_q(set_queue_name, logger, custodian=False):
    q_dir, q_name = os.path.split(set_queue_name)

    q = None
    try: 
	# so that the set.queue can be rwx by all. this allows a root
	# process to consume the set.queue while any user enqueues to it.
	os.umask(0)
	q = work_q.Queue(q_dir, q_name, custodian=custodian, 
			 logger=logger, mode=0777)

    except work_q.Error, errval:
	if errval: logger.log(str(errval))
	del q
	q = None

    return q

def init_dqm(config_obj):

    # Sorry about the syntax. This next line takes a list of 
    # of host:port strings and converts it to a list of tuples
    # in which the first entry is the host and the second entry is a 
    # numeric value for the port.
    portal_tuples = [(x.split(':')[0], int(x.split(':')[1])) \
		     for x in config_obj['portal']]
    return dqm_client.DQM(portal_tuples)

class Receiver:

    MAX_TRIES        = 5
    ENQUEUE_INTERVAL = 15

    def __init__(self, config_obj, logger=None):
	
	if logger is None: logger = dqm.logger.stderr_logger()
	self.logger     = logger

	self.config_obj = config_obj
	self.signers    = config_obj['trusted_signer']
	self.set_q_name = config_obj['set_queue']
	self.dqm_client = init_dqm(config_obj)
	self.set_q      = None

    def process_stamp(self, stamp_str):
	"""This method checks a stamp for validity. If the stamp is valid
	and wasn't previously in the DQM, this method queues it to a work
	queue, the consumer of which cancels the stamps in the DQM. Returns an 
	error defined in dqm/err.py"""

	rc = dqm_stamp.SimpleStamp().verify(stamp_str, self.signers)
	if rc != dqm_err.SUCCESS:
	    msg = "verification error: %s. stamp: %s" % \
		  (dqm_err.str_codes[rc], stamp_str)
	    self.logger.log(msg)
	    return rc
	
	val = sha.new(stamp_str).digest()
	key = sha.new(val).digest()
  
	self.dqm_client.test(key, val, self.test_cb)

	# We invoke the event loop here because it prevents the
	# rest of the code from having to become asynchronous. This is the
	# equivalent of doing the call synchronously.
	self.res = None
	dqm_events.loop(False)

	assert(self.res is not None)
	in_dqm, err = self.res

	if err != '':
	    msg = "'test' failed. stamp: %s. err: %s" % (stamp_str, err)
	    self.logger.log(msg)

	if in_dqm is True: 
	    rc = dqm_err.ALREADY_CANCELLED
	else:              
	    rc = dqm_err.SUCCESS
	    self.cancel_stamp(key, val)

	return rc

    def test_cb(self, res):

	# pick up the return value here; interpret it later
	self.res = res
	dqm_events.stop_looping()

    # this function takes a (key,val) pair and enqueues it to an on-disk
    # queue that is read by the stamp-canceller program.
    def cancel_stamp(self, key, val):
  
	tries = 0
	while tries < self.MAX_TRIES:

	    try: 
		if self.set_q is None:
		    self.set_q = init_set_q(self.set_q_name, self.logger)

		if self.set_q is not None:
		    rc = self.set_q.enqueue(key, val, block=False)
		else:
		    self.logger.log("couldn't initialize DB in cancel_stamp")
		    raise work_q.Error()
		break

	    except work_q.Error, errval:
		tries = tries + 1
		self.logger.log(str(errval))

		# keep trying until the stamp-canceller notices that the DB is 
		# corrupted and starts up again to do recovery
		if self.set_q is not None: 
		    self.set_q.close()
		    del self.set_q
		    self.set_q = None

		time.sleep(self.ENQUEUE_INTERVAL)
		continue 

	if rc: self.logger.log("Enqueue succesful on try " + str(tries + 1))
	else:  self.logger.log("Enqueue command failed")
