/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include "tt.h"

inline bool
aligned(char* p, u_int32_t align)
{
    return !(reinterpret_cast<unsigned int>(p) & (align - 1));
}

inline char*
align_down(char* p, u_int32_t align)
{   
    return reinterpret_cast<char*>(
               (reinterpret_cast<unsigned int>(p) & ~(align - 1)));
}

inline char*
align_up(char* p, u_int32_t align)
{
    if (!aligned(p, align)) 
	return align_down(
	           reinterpret_cast<char*>(
		      reinterpret_cast<unsigned int>(p) + align), align);
    return p;
}

typedef ssize_t (*io_func)(int, void*, size_t);
typedef ssize_t (*io_func_ra)(int, void*, size_t, off_t);

void
worker_sequential(int fd, char* buf, int block_size, io_func op)
{
    int rc;
    tt t;
    time_it s;
    u_int32_t i = 0;

    for (i = 0; ; i++) {

	s.begin();
	if ((rc = op(fd, buf, block_size)) != block_size) {
	    printf("io error: %d\n", rc);
	    exit(1);
	}
	s.end();

	t.inc();

	if ( (i % 250) == 0) 
	    printf("ops per sec: %.2f. avg op: %.2f. min: %.2f\n", 
		   t.num_per_sec(), s.avg_us(), s.min_us());
    }
}

void
worker_rand_access(int fd, char* buf, int block_size, io_func_ra op)
{
    int rc;
    tt t;
    time_it s;
    u_int32_t i = 0;
    u_int32_t r;
    time_t tm;
    off_t off;

    time(&tm);
    srand((unsigned int)tm);

    for (i = 0; ; i++) {

	r = rand();
	off = (r + (block_size - 1)) & ~(block_size - 1);

	s.begin();
	if ((rc = op(fd, buf, block_size, off)) != block_size) {
	    printf("io error: %d\n", rc);
	    exit(1);
	}
	s.end();

	t.inc();

	if ( (i % 250) == 0) 
	    printf("ops per sec: %.2f. avg op: %.2f. min: %.2f\n", 
		   t.num_per_sec(), s.avg_us(), s.min_us());
    }

}

void
usage(char** argv)
{
    printf("usage: %s [-d] [-r] dev blocksize r|w\n", argv[0]);
    printf("  -d: turn on O_DIRECT flag\n");
    printf("  -r: random access (default is sequential)\n");
    exit(1);
}

int main(int argc, char** argv)
{
    // required positional arguments
    int off_filename  = 0;
    int off_blocksize = 1;
    int off_op        = 2;

    u_int32_t flags = (O_CREAT | O_RDWR);
    bool rand_access = false;
    bool direct = false;
    int fd;

    int opt;
    opterr = 0;
    while ( (opt = getopt(argc, argv, "dr")) != -1) {

	switch (opt) {

	    case 'd':
		direct = true;
		break;

	    case 'r':
		rand_access = true;
		break;

	    default:
		usage(argv);
	}
    }

    if (optind != argc - 3)
	usage(argv);

    int ind_fname  = optind + off_filename;
    int ind_blk_sz = optind + off_blocksize;
    int ind_op     = optind + off_op;

    char* fname  = argv[ind_fname];
    int   blk_sz = atoi(argv[ind_blk_sz]);
    io_func op = NULL;

    switch (argv[ind_op][0]) {
	case 'r':
	case 'R':
	    op = read;
	    break;

	case 'w':
	case 'W':
	    op = reinterpret_cast<io_func>(write);
	    break;
	default:
	    usage(argv);
    }

    // linux requires 512 byte alignment
    char* buf = new char[blk_sz + 512];
    char* buf_a = align_up(buf, 512);

    if (direct) 
	flags |= O_DIRECT;

    if ((fd = open(fname, flags, 0644)) < 0) {
	printf("open failed. error: %s\n", strerror(errno));
	return 1;
    }

    if (rand_access) {

	io_func_ra ra_op = NULL;
	if (op == read)       
	    ra_op = pread;
	else if (op == reinterpret_cast<io_func>(write)) 
	    ra_op = reinterpret_cast<io_func_ra>(pwrite);
	else assert(false);

	worker_rand_access(fd, buf_a, blk_sz, ra_op);
    } else {	
	worker_sequential(fd, buf_a, blk_sz, op);
    }
}
