/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include "index_map.h"

const index_t index_map::EMPTY_ENTRY = 0xffffffff;

// the purpose of hashseed is to make the probe sequence
// unpredictable to an adversary. otherwise, the adversary can mess
// up the "randomness" on which the data structure depends.

inline u_int32_t
index_map::get_loc(ptr<dqm_key> k)
{
    return (sdbm_hash_bytes(k->base(), k->size(), hashseed1) % num_buckets);
}

inline u_int8_t 
index_map::EXTRACT_CKSUM(entry_t entry)
{
    u_int8_t ret;
    u_int32_t ret32;
    ret32 = (0xff000000 & entry) >> 24;
    ret = static_cast<u_int8_t>(ret32); 
    return ret;
}

inline index_t 
index_map::EXTRACT_INDEX(entry_t entry)
{
    index_t ret;
    ret = (0x00ffffff & entry);
    return ret; 
}

inline index_map::entry_t
index_map::COMBINE(u_int8_t ck, index_t index)
{
    u_int32_t i = static_cast<u_int32_t>(index);
    assert( !(i & 0xff000000) );
    return ( (ck << 24) | (i & 0x00ffffff) );

}

index_t 
index_map::lookup(ptr<dqm_key> k)
{
    index_t* idxp;

    if ((idxp = overflow[k]) != NULL)  
	return *idxp;

    u_int32_t i;
    u_int8_t  ck  = cksum(k);
    u_int32_t loc = get_loc(k);

    for (i = 1; i <= num_buckets; i++) {

	if (tab[loc] == static_cast<entry_t>(EMPTY_ENTRY)) 
	    return EMPTY_ENTRY;
	else if (EXTRACT_CKSUM(tab[loc]) == ck) 
	    return EXTRACT_INDEX(tab[loc]);

	// this is quadratic probing. these go in the sequence
	// l, (l + 1), (l + 3), (l + 6), which is:
	// l, (l + (2 2)), (l + (3 2)), (l + (4 2)), ..., (l + (n 2)), where
	//  (a b) is \binom{a}{b}
	loc = (loc + i) % num_buckets;
    }

    assert(i <= num_buckets);
    return EMPTY_ENTRY;
}

// Assumption is being made here, namely that every 'k' is novel. It is
// the caller's responsibility to make sure that 'k' does not already
// live in the map. If 'k' is inserted multiple times, the second time
// will definitely cause 'k' to be inserted into our internal overflow
// table. 
void
index_map::insert(ptr<dqm_key> k, index_t idx)
{
    if (overflow[k] != NULL)
	return;

    u_int32_t i;
    u_int8_t  ck  = cksum(k);
    u_int32_t loc = get_loc(k);

    for (i = 1; i <= num_buckets; i++) {

	if (tab[loc] == static_cast<entry_t>(EMPTY_ENTRY)) {

	    tab[loc] = COMBINE(ck, idx);
	    return;

	} else if (EXTRACT_CKSUM(tab[loc]) == ck) {

	    // we assume that we're not seeing the same key twice.
	    overflow.insert(k, idx);
	    ++overflows;
	    return;
	}
	loc = (loc + i) % num_buckets;
    }
    assert(i <= num_buckets);
}

u_int8_t
index_map::cksum(ptr<dqm_key> key)
{
    u_int32_t hashval = hash_bytes(key->base(), key->size(), hashseed2);

    return (hashval & 0x000000ff);
}

