/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include <math.h>
#include "sha1.h"
#include "nodetable_geom.h"
#include "dqm_types.h"

nodetable_geom::nodetable_geom(double dee, int num_rings) : 
  d(dee), nrings(num_rings), nnodes(0)
{
    memset(&local_id, 0, sizeof(local_id));

    double log_d_modulus = log(MODULUS) / log (d);
    u_int32_t max_fingers = (u_int32_t)(log_d_modulus + 1);

    // pre-compute, for efficiency
    for (u_int32_t i = 0; i <= max_fingers; i++) 
	powers_of_d.push_back(pow(d, i));

    for (u_int32_t j = 0; j < nrings; j++) 
	rings.push_back(New ring_lst);

}

nodetable_geom::~nodetable_geom()
{
    ringmember_lstelm* curr;
    ring_lst* ring;

    for (u_int32_t i = 0; i < nrings; i++) {
	ring = rings[i];
	for (curr = ring->first(); curr; curr = ring->first()) {
	    ring->remove(curr->alias);
	    delete curr;
	}
	delete ring;
    }
}

// Calculates the path from the current id to the key, in one of
// the virtual rings. This function makes the decision about which
// ring to use and in fact encapsulates all of that logic: callers
// simply supply a key and get back a list of (IP,port) pairs. If the
// nodes that would have been up are down, then those nodes are
// simply absent from the list.  If a blank list is returned here,
// then either no nodes along the path are up or the current node is
// the successor. It is an error to call this function without having
// called set_local_node
ptr<node_list> 
nodetable_geom::calc_path(ptr<dqm_key> key)
{
    assert(stopping_ratio != -1.0);

    ptr<node_list> ret = New refcounted<node_list>;
    ringmember_lstelm *root_p, *succ_p;
    u_int32_t X, i = 1, target;
    double dist_from_end;

    node_alias last_alias_added = 0;
    node_alias succ_alias;
    node_id    succ_id;

    u_int32_t  which_ring    = choose_ring(*key);
    node_alias key_as_alias  
                 = hash2alias(reinterpret_cast<u_char*>(key->base()));

    root_p = get_succ(key_as_alias, which_ring);
    node_alias root_alias = root_p->alias;
    node_id    root_id    = root_p->id;

    alias_vec* local_aliases = node_aliases[local_id];
    node_alias our_alias     = (*local_aliases)[which_ring];

    // we're the successor. chance has smiled on us, and let's not push it.
    if (our_alias == root_alias)
	return ret;

    // this says: let X = (root_alias - our_alias) % MODULUS
    X = (root_alias < our_alias) 
          ? MODULUS + root_alias - our_alias
	  : root_alias - our_alias;
			    
    while (true) {

	// division is floating point b/c powers_of_d holds doubles
	dist_from_end = X / powers_of_d[i];

	// this says: let target = (root_alias - X/d^i) % MODULUS
	target = (u_int32_t)(
	          (root_alias < dist_from_end) 
	            ? MODULUS + root_alias - dist_from_end
		    : root_alias - dist_from_end);
	
	succ_p = get_succ(target, which_ring);
	succ_alias = succ_p->alias;
	succ_id    = succ_p->id;

	// Add the node if it's up and (EITHER we haven't added
	// any nodes OR this node is not already on our 'to-return' list)
	if (*nodes[succ_id] == UP && 
	   (ret->size() == 0 || succ_alias != last_alias_added)) {
	     ret->push_back(succ_id);
	     last_alias_added = succ_alias;
	}

	// We got the root, which means we're done. no other nodes
	// in the interval [root - X/d^i,root]. 
	if (succ_alias == root_alias)
	    break;

	if (dist_from_end < stopping_ratio) {
	    // karger's optimization: we're within d/M of the root,
	    // so jump to the root, and get out of here.
	    if (*nodes[root_id] == UP) {
		// 'course, we only add the root if we think it's UP.
		ret->push_back(root_id);
	    }
	    break;
	}

	++i;
    }

    return ret;
}

void
nodetable_geom::set_local_node(const node_id& id)
{
    local_id = id;
}

void
nodetable_geom::add_node(const node_id& id, const str& alias_seed)
{
    alias_vec aliases = get_node_aliases(alias_seed, nrings);
    ring_lst* ring;
    node_alias alias;
    static int dup_cnt = 0;

    for (u_int32_t i = 0; i < nrings; i++) {

	ring  = rings[i];
	alias = aliases[i];
   
	ringmember_lstelm* to_insert = New ringmember_lstelm;
	to_insert->alias = alias;
	to_insert->id    = id;
	
	if (!ring->insert(to_insert)) {
	    ++dup_cnt;
	    warnx("duplicate id. %d duplicates so far.\n", dup_cnt);
	}
    }

    node_aliases.insert(id, aliases);
    nodes.insert(id, DOWN);
    ++nnodes;

    // the stopping rule is X / d^i < d / M. 
    //  M just changed, so update the stopping ratio (the thing on the
    //  right-hand side of the inequality.
    stopping_ratio = d / nnodes;
}

void
nodetable_geom::remove_node(const node_id& id)
{
    alias_vec* aliases = node_aliases[id];
    assert(aliases);

    ring_lst* ring;
    node_alias alias;

    for (u_int32_t i = 0; i < nrings; i++) {

	ring  = rings[i];
	alias = (*aliases)[i];

	ringmember_lstelm* curr;

	curr = ring->remove(alias);
	delete curr;
    }

    nodes.remove(id);
    node_aliases.remove(id);

    --nnodes;
    stopping_ratio = (nnodes == 0) ? -1.0 : (double)d / nnodes;
}

void
nodetable_geom::dump()
{
    dump_nodes_and_aliases(&nodes, &node_aliases);
   
    warnx("\n");

    // print rings
    for (u_int32_t i = 0; i < nrings; i++) {
	ring_lst* ring = rings[i];
	ringmember_lstelm* curr;

	warnx("ring %d: ", i);

	// all the aliases for a given node
	for (curr = ring->first(); curr; curr = ring->next(curr)) 
	    warnx("%.6x ", curr->alias);

	warnx("\n");
    }
}

// returns the id of the node that succeeds the node with alias 'alias'
// in ring 'which_ring'.
inline nodetable::ringmember_lstelm* 
nodetable_geom::get_succ(node_alias alias, u_int32_t which_ring)
{
    ring_lst* ring = rings[which_ring];
    return nodetable::get_succ(alias, ring);
}

// We use bytes 4,5,6 of the key and take this number modulo the number
// of rings.
inline u_int32_t
nodetable_geom::choose_ring(const dqm_key& key)
{
    const char* key_bytes = key.base();
    u_int32_t bottom_bits = ((key_bytes[4] << 16) & 0x00ff0000) | 
			    ((key_bytes[5] << 8)  & 0x0000ff00) |
			    (key_bytes[6] & 0x000000ff);
    u_int32_t ring = bottom_bits % nrings;
    return ring;
}

// For now, use the top 3 bytes of the hash for the purposes of
// calculating successors and the lookup path.
nodetable::node_alias 
nodetable_geom::hash2alias(const u_char* hash) {
    return ((hash[0] << 16) & 0x00ff0000) | 
	   ((hash[1] << 8)  & 0x0000ff00) | 
	   (hash[2] & 0x000000ff);
}


